"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var rollup = __importStar(require("rollup"));
var chalk_1 = __importDefault(require("chalk"));
var pretty_ms_1 = __importDefault(require("pretty-ms"));
var logging_1 = require("../logging");
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var promise_1 = require("../../../src/utils/promise");
var sourceMappingUrl_1 = __importDefault(require("../sourceMappingUrl"));
function build(inputOptions, outputOptions, warnings, silent) {
    if (silent === void 0) { silent = false; }
    var useStdout = outputOptions.length === 1 && !outputOptions[0].file && inputOptions.input instanceof Array === false;
    var start = Date.now();
    var files = useStdout
        ? ['stdout']
        : outputOptions.map(function (t) { return relativeId_1.default(t.file || t.dir); });
    if (!silent)
        logging_1.stderr(chalk_1.default.cyan("\n" + chalk_1.default.bold(typeof inputOptions.input === 'string' ? inputOptions.input : inputOptions.input && inputOptions.input.join(', ')) + " \u2192 " + chalk_1.default.bold(files.join(', ')) + "..."));
    return rollup
        .rollup(inputOptions)
        .then(function (bundle) {
        if (useStdout) {
            var output_1 = outputOptions[0];
            if (output_1.sourcemap && output_1.sourcemap !== 'inline') {
                logging_1.handleError({
                    code: 'MISSING_OUTPUT_OPTION',
                    message: 'You must specify an --output (-o) option when creating a file with a sourcemap'
                });
            }
            return bundle.generate(output_1).then(function (_a) {
                var code = _a.code, map = _a.map;
                if (output_1.sourcemap === 'inline') {
                    code += "\n//# " + sourceMappingUrl_1.default + "=" + map.toUrl() + "\n";
                }
                process.stdout.write(code);
            });
        }
        return promise_1.mapSequence(outputOptions, function (output) {
            return bundle.write(output);
        });
    })
        .then(function () {
        warnings.flush();
        if (!silent)
            logging_1.stderr(chalk_1.default.green("created " + chalk_1.default.bold(files.join(', ')) + " in " + chalk_1.default.bold(pretty_ms_1.default(Date.now() - start))));
    })
        .catch(logging_1.handleError);
}
exports.default = build;
