"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs_1 = require("fs");
var require_relative_1 = __importDefault(require("require-relative"));
var logging_1 = require("../logging");
var mergeOptions_1 = __importDefault(require("../../../src/utils/mergeOptions"));
var batchWarnings_1 = __importDefault(require("./batchWarnings"));
var loadConfigFile_1 = __importDefault(require("./loadConfigFile"));
var sequence_1 = __importDefault(require("../utils/sequence"));
var build_1 = __importDefault(require("./build"));
var watch_1 = __importDefault(require("./watch"));
function runRollup(command) {
    if (command._.length >= 1) {
        if (command.input) {
            logging_1.handleError({
                code: 'DUPLICATE_IMPORT_OPTIONS',
                message: 'use --input, or pass input path as argument'
            });
        }
    }
    if (command.output && command.output.dir) {
        command.input = command._;
        command._ = [];
    }
    else if (command._.length === 1) {
        command.input = command._[0];
    }
    if (command.environment) {
        var environment = Array.isArray(command.environment)
            ? command.environment
            : [command.environment];
        environment.forEach(function (arg) {
            arg.split(',').forEach(function (pair) {
                var _a = pair.split(':'), key = _a[0], value = _a[1];
                if (value) {
                    process.env[key] = value;
                }
                else {
                    process.env[key] = String(true);
                }
            });
        });
    }
    var configFile = command.config === true ? 'rollup.config.js' : command.config;
    if (configFile) {
        if (configFile.slice(0, 5) === 'node:') {
            var pkgName = configFile.slice(5);
            try {
                configFile = require_relative_1.default.resolve("rollup-config-" + pkgName, process.cwd());
            }
            catch (err) {
                try {
                    configFile = require_relative_1.default.resolve(pkgName, process.cwd());
                }
                catch (err) {
                    if (err.code === 'MODULE_NOT_FOUND') {
                        logging_1.handleError({
                            code: 'MISSING_EXTERNAL_CONFIG',
                            message: "Could not resolve config file " + configFile
                        });
                    }
                    throw err;
                }
            }
        }
        else {
            // find real path of config so it matches what Node provides to callbacks in require.extensions
            configFile = fs_1.realpathSync(configFile);
        }
        if (command.watch)
            process.env.ROLLUP_WATCH = 'true';
        loadConfigFile_1.default(configFile, command.silent)
            .then(function (normalized) { return execute(configFile, normalized, command); })
            .catch(logging_1.handleError);
    }
    else {
        return execute(configFile, [{ input: null }], command);
    }
}
exports.default = runRollup;
function execute(configFile, configs, command) {
    if (command.watch) {
        watch_1.default(configFile, configs, command, command.silent);
    }
    else {
        return sequence_1.default(configs, function (config) {
            var warnings = batchWarnings_1.default();
            var _a = mergeOptions_1.default({ config: config, command: command, defaultOnWarnHandler: warnings.add }), inputOptions = _a.inputOptions, outputOptions = _a.outputOptions, deprecations = _a.deprecations, optionError = _a.optionError;
            if (deprecations.length) {
                inputOptions.onwarn({
                    code: 'DEPRECATED_OPTIONS',
                    message: "The following options have been renamed \u2014 please update your config: " + deprecations
                        .map(function (option) { return option.old + " -> " + option.new; })
                        .join(', '),
                    deprecations: deprecations
                });
            }
            if (optionError)
                inputOptions.onwarn({ code: 'UNKNOWN_OPTION', message: optionError });
            return build_1.default(inputOptions, outputOptions, warnings, command.silent);
        });
    }
}
