"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var chalk_1 = __importDefault(require("chalk"));
var rollup = __importStar(require("rollup"));
var batchWarnings_1 = __importDefault(require("./batchWarnings"));
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var logging_1 = require("../logging");
function loadConfigFile(configFile, silent) {
    if (silent === void 0) { silent = false; }
    var warnings = batchWarnings_1.default();
    return rollup
        .rollup({
        input: configFile,
        external: function (id) {
            return ((id[0] !== '.' && !path_1.default.isAbsolute(id)) ||
                id.slice(-5, id.length) === '.json');
        },
        onwarn: warnings.add,
    })
        .then(function (bundle) {
        if (!silent && warnings.count > 0) {
            logging_1.stderr(chalk_1.default.bold("loaded " + relativeId_1.default(configFile) + " with warnings"));
            warnings.flush();
        }
        return bundle.generate({
            format: 'cjs'
        });
    })
        .then(function (_a) {
        var code = _a.code;
        // temporarily override require
        var defaultLoader = require.extensions['.js'];
        require.extensions['.js'] = function (m, filename) {
            if (filename === configFile) {
                m._compile(code, filename);
            }
            else {
                defaultLoader(m, filename);
            }
        };
        delete require.cache[configFile];
        return Promise.resolve(require(configFile)).then(function (configs) {
            if (Object.keys(configs).length === 0) {
                logging_1.handleError({
                    code: 'MISSING_CONFIG',
                    message: 'Config file must export an options object, or an array of options objects',
                    url: 'https://rollupjs.org/#using-config-files'
                });
            }
            require.extensions['.js'] = defaultLoader;
            return Array.isArray(configs) ? configs : [configs];
        });
    });
}
exports.default = loadConfigFile;
