"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs_1 = __importDefault(require("fs"));
var rollup = __importStar(require("rollup"));
var chalk_1 = __importDefault(require("chalk"));
var pretty_ms_1 = __importDefault(require("pretty-ms"));
var signal_exit_1 = __importDefault(require("signal-exit"));
var date_time_1 = __importDefault(require("date-time"));
var mergeOptions_1 = __importDefault(require("../../../src/utils/mergeOptions"));
var batchWarnings_1 = __importDefault(require("./batchWarnings"));
var alternateScreen_1 = __importDefault(require("./alternateScreen"));
var loadConfigFile_1 = __importDefault(require("./loadConfigFile"));
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var logging_1 = require("../logging");
;
function watch(configFile, configs, command, silent) {
    if (silent === void 0) { silent = false; }
    var isTTY = Boolean(process.stderr.isTTY);
    var screen = alternateScreen_1.default(isTTY);
    screen.open();
    var warnings = batchWarnings_1.default();
    var watcher;
    var configWatcher;
    function start(configs) {
        screen.reset(chalk_1.default.underline("rollup v" + rollup.VERSION));
        var screenWriter = screen.reset;
        configs = configs.map(function (options) {
            var merged = mergeOptions_1.default({ config: options, command: command, defaultOnWarnHandler: warnings.add });
            var result = Object.assign({}, merged.inputOptions, {
                output: merged.outputOptions
            });
            if (merged.deprecations.length) {
                if (!result.watch)
                    result.watch = {};
                result.watch._deprecations = merged.deprecations;
            }
            if (merged.optionError)
                merged.inputOptions.onwarn({ message: merged.optionError, code: 'UNKNOWN_OPTION' });
            if (merged.inputOptions.watch &&
                merged.inputOptions.watch.clearScreen === false) {
                screenWriter = logging_1.stderr;
            }
            return result;
        });
        watcher = rollup.watch(configs);
        watcher.on('event', function (event) {
            switch (event.code) {
                case 'FATAL':
                    screen.close();
                    logging_1.handleError(event.error, true);
                    process.exit(1);
                    break;
                case 'ERROR':
                    warnings.flush();
                    logging_1.handleError(event.error, true);
                    break;
                case 'START':
                    screenWriter(chalk_1.default.underline("rollup v" + rollup.VERSION));
                    break;
                case 'BUNDLE_START':
                    if (!silent)
                        logging_1.stderr(chalk_1.default.cyan("bundles " + chalk_1.default.bold(typeof event.input === 'string' ? event.input : event.input.join(', ')) + " \u2192 " + chalk_1.default.bold(event.output.map(relativeId_1.default).join(', ')) + "..."));
                    break;
                case 'BUNDLE_END':
                    warnings.flush();
                    if (!silent)
                        logging_1.stderr(chalk_1.default.green("created " + chalk_1.default.bold(event.output.map(relativeId_1.default).join(', ')) + " in " + chalk_1.default.bold(pretty_ms_1.default(event.duration))));
                    break;
                case 'END':
                    if (!silent && isTTY) {
                        logging_1.stderr("\n[" + date_time_1.default() + "] waiting for changes...");
                    }
            }
        });
    }
    // catch ctrl+c, kill, and uncaught errors
    var removeOnExit = signal_exit_1.default(close);
    process.on('uncaughtException', close);
    // only listen to stdin if it is a pipe
    if (!process.stdin.isTTY) {
        process.stdin.on('end', close); // in case we ever support stdin!
        process.stdin.resume();
    }
    function close(err) {
        if (err) {
            console.error(err);
        }
        removeOnExit();
        process.removeListener('uncaughtException', close);
        // removing a non-existent listener is a no-op
        process.stdin.removeListener('end', close);
        screen.close();
        if (watcher)
            watcher.close();
        if (configWatcher)
            configWatcher.close();
        if (err) {
            process.exit(1);
        }
    }
    start(configs);
    if (configFile && !configFile.startsWith('node:')) {
        var restarting_1 = false;
        var aborted_1 = false;
        var configFileData_1 = fs_1.default.readFileSync(configFile, 'utf-8');
        var restart_1 = function () {
            var newConfigFileData = fs_1.default.readFileSync(configFile, 'utf-8');
            if (newConfigFileData === configFileData_1)
                return;
            configFileData_1 = newConfigFileData;
            if (restarting_1) {
                aborted_1 = true;
                return;
            }
            restarting_1 = true;
            loadConfigFile_1.default(configFile, silent)
                .then(function (configs) {
                restarting_1 = false;
                if (aborted_1) {
                    aborted_1 = false;
                    restart_1();
                }
                else {
                    watcher.close();
                    start(configs);
                }
            })
                .catch(function (err) {
                logging_1.handleError(err, true);
            });
        };
        configWatcher = fs_1.default.watch(configFile, function (event) {
            if (event === 'change')
                restart_1();
        });
    }
}
exports.default = watch;
