"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var Node_1 = require("./shared/Node");
var ClassDeclaration_1 = require("./ClassDeclaration");
var FunctionDeclaration_1 = require("./FunctionDeclaration");
var renderHelpers_1 = require("../../utils/renderHelpers");
var ObjectExpression_1 = require("./ObjectExpression");
var object_1 = require("../../utils/object");
var WHITESPACE = /\s/;
// The header ends at the first non-white-space after "default"
function getDeclarationStart(code, start) {
    if (start === void 0) { start = 0; }
    start = renderHelpers_1.findFirstOccurrenceOutsideComment(code, 'default', start) + 7;
    while (WHITESPACE.test(code[start]))
        start++;
    return start;
}
function getIdInsertPosition(code, declarationKeyword, start) {
    if (start === void 0) { start = 0; }
    var declarationEnd = renderHelpers_1.findFirstOccurrenceOutsideComment(code, declarationKeyword, start) + declarationKeyword.length;
    code = code.slice(declarationEnd, renderHelpers_1.findFirstOccurrenceOutsideComment(code, '{', declarationEnd));
    var generatorStarPos = renderHelpers_1.findFirstOccurrenceOutsideComment(code, '*');
    if (generatorStarPos === -1) {
        return declarationEnd;
    }
    return declarationEnd + generatorStarPos + 1;
}
var needsToBeWrapped = ObjectExpression_1.isObjectExpression;
var ExportDefaultDeclaration = /** @class */ (function (_super) {
    __extends(ExportDefaultDeclaration, _super);
    function ExportDefaultDeclaration() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    ExportDefaultDeclaration.prototype.bindNode = function () {
        if (this.declarationName) {
            this.variable.setOriginalVariable(this.scope.findVariable(this.declarationName));
        }
    };
    ExportDefaultDeclaration.prototype.initialiseNode = function () {
        this.declarationName =
            (this.declaration.id && this.declaration.id.name) ||
                this.declaration.name;
        this.variable = this.scope.addExportDefaultDeclaration(this.declarationName || this.module.basename(), this);
    };
    ExportDefaultDeclaration.prototype.render = function (code, options, _a) {
        var _b = _a === void 0 ? object_1.BLANK : _a, start = _b.start, end = _b.end;
        var declarationStart = getDeclarationStart(code.original, this.start);
        if (FunctionDeclaration_1.isFunctionDeclaration(this.declaration)) {
            this.renderNamedDeclaration(code, declarationStart, 'function', this.declaration.id === null, options);
        }
        else if (ClassDeclaration_1.isClassDeclaration(this.declaration)) {
            this.renderNamedDeclaration(code, declarationStart, 'class', this.declaration.id === null, options);
        }
        else if (this.variable.referencesOriginal()) {
            // Remove altogether to prevent re-declaring the same variable
            if (!options.systemBindings) {
                code.remove(start, end);
            }
            else {
                code.overwrite(start, end, "exports('" + this.variable.exportName + "', " + this.variable.getName() + ");");
            }
            return;
        }
        else if (this.variable.included) {
            this.renderVariableDeclaration(code, declarationStart, options);
        }
        else {
            this.renderForSideEffectsOnly(code, declarationStart);
        }
        _super.prototype.render.call(this, code, options);
    };
    ExportDefaultDeclaration.prototype.renderNamedDeclaration = function (code, declarationStart, declarationKeyword, needsId, options) {
        var name = this.variable.getName();
        // Remove `export default`
        code.remove(this.start, declarationStart);
        if (needsId) {
            code.appendLeft(getIdInsertPosition(code.original, declarationKeyword, declarationStart), " " + name);
        }
        if (options.systemBindings && ClassDeclaration_1.isClassDeclaration(this.declaration)) {
            code.appendRight(this.end, " exports('default', " + name + ");");
        }
    };
    ExportDefaultDeclaration.prototype.renderVariableDeclaration = function (code, declarationStart, options) {
        var systemBinding = options.systemBindings ? "exports('" + this.variable.exportName + "', " : '';
        code.overwrite(this.start, declarationStart, this.module.graph.varOrConst + " " + this.variable.getName() + " = " + systemBinding);
        if (systemBinding) {
            code.prependRight(this.end - 1, ')');
        }
    };
    ExportDefaultDeclaration.prototype.renderForSideEffectsOnly = function (code, declarationStart) {
        code.remove(this.start, declarationStart);
        if (needsToBeWrapped(this.declaration)) {
            code.appendLeft(declarationStart, '(');
            if (code.original[this.end - 1] === ';') {
                code.prependRight(this.end - 1, ')');
            }
            else {
                code.prependRight(this.end, ');');
            }
        }
    };
    return ExportDefaultDeclaration;
}(Node_1.NodeBase));
exports.default = ExportDefaultDeclaration;
ExportDefaultDeclaration.prototype.needsBoundaries = true;
ExportDefaultDeclaration.prototype.isExportDeclaration = true;
