"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var object_1 = require("../utils/object");
var CallOptions_1 = __importDefault(require("./CallOptions"));
var VariableReassignmentTracker_1 = require("./variables/VariableReassignmentTracker");
function assembleMemberDescriptions(memberDescriptions, inheritedDescriptions) {
    if (inheritedDescriptions === void 0) { inheritedDescriptions = null; }
    return Object.create(inheritedDescriptions, memberDescriptions);
}
exports.UNKNOWN_VALUE = { toString: function () { return '[[UNKNOWN]]'; } };
exports.UNKNOWN_EXPRESSION = {
    reassignPath: function () { },
    forEachReturnExpressionWhenCalledAtPath: function () { },
    getValue: function () { return exports.UNKNOWN_VALUE; },
    hasEffectsWhenAccessedAtPath: function (path) { return path.length > 0; },
    hasEffectsWhenAssignedAtPath: function (path) { return path.length > 0; },
    hasEffectsWhenCalledAtPath: function () { return true; },
    someReturnExpressionWhenCalledAtPath: function () { return true; },
    toString: function () { return '[[UNKNOWN]]'; }
};
var returnsUnknown = { value: { returns: exports.UNKNOWN_EXPRESSION, callsArgs: null } };
var callsArgReturnsUnknown = { value: { returns: exports.UNKNOWN_EXPRESSION, callsArgs: [0] } };
exports.UNKNOWN_ARRAY_EXPRESSION = {
    reassignPath: function () { },
    forEachReturnExpressionWhenCalledAtPath: function () { },
    getValue: function () { return exports.UNKNOWN_VALUE; },
    hasEffectsWhenAccessedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenAssignedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenCalledAtPath: function (path, callOptions, options) {
        if (path.length === 1) {
            return hasMemberEffectWhenCalled(exports.arrayMembers, path[0], callOptions, options);
        }
        return true;
    },
    someReturnExpressionWhenCalledAtPath: function (path, callOptions, predicateFunction, options) {
        if (path.length === 1) {
            return someMemberReturnExpressionWhenCalled(exports.arrayMembers, path[0], callOptions, predicateFunction, options);
        }
        return true;
    },
    toString: function () { return '[[UNKNOWN ARRAY]]'; }
};
var returnsArray = { value: { returns: exports.UNKNOWN_ARRAY_EXPRESSION, callsArgs: null } };
var callsArgReturnsArray = { value: { returns: exports.UNKNOWN_ARRAY_EXPRESSION, callsArgs: [0] } };
var UNKNOWN_LITERAL_BOOLEAN = {
    reassignPath: function () { },
    forEachReturnExpressionWhenCalledAtPath: function () { },
    getValue: function () { return exports.UNKNOWN_VALUE; },
    hasEffectsWhenAccessedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenAssignedAtPath: function (path) { return path.length > 0; },
    hasEffectsWhenCalledAtPath: function (path) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath) || !literalBooleanMembers[subPath];
        }
        return true;
    },
    someReturnExpressionWhenCalledAtPath: function (path, _callOptions, predicateFunction, options) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath)
                || !literalBooleanMembers[subPath]
                || predicateFunction(options)(literalBooleanMembers[subPath].returns);
        }
        return true;
    },
    toString: function () { return '[[UNKNOWN BOOLEAN]]'; }
};
var returnsBoolean = { value: { returns: UNKNOWN_LITERAL_BOOLEAN, callsArgs: null } };
var callsArgReturnsBoolean = { value: { returns: UNKNOWN_LITERAL_BOOLEAN, callsArgs: [0] } };
var UNKNOWN_LITERAL_NUMBER = {
    reassignPath: function () { },
    forEachReturnExpressionWhenCalledAtPath: function () { },
    getValue: function () { return exports.UNKNOWN_VALUE; },
    hasEffectsWhenAccessedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenAssignedAtPath: function (path) { return path.length > 0; },
    hasEffectsWhenCalledAtPath: function (path) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath) || !literalNumberMembers[subPath];
        }
        return true;
    },
    someReturnExpressionWhenCalledAtPath: function (path, _callOptions, predicateFunction, options) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath)
                || !literalNumberMembers[subPath]
                || predicateFunction(options)(literalNumberMembers[subPath].returns);
        }
        return true;
    },
    toString: function () { return '[[UNKNOWN NUMBER]]'; }
};
var returnsNumber = { value: { returns: UNKNOWN_LITERAL_NUMBER, callsArgs: null } };
var callsArgReturnsNumber = { value: { returns: UNKNOWN_LITERAL_NUMBER, callsArgs: [0] } };
var UNKNOWN_LITERAL_STRING = {
    reassignPath: function () { },
    forEachReturnExpressionWhenCalledAtPath: function () { },
    getValue: function () { return exports.UNKNOWN_VALUE; },
    hasEffectsWhenAccessedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenAssignedAtPath: function (path) { return path.length > 0; },
    hasEffectsWhenCalledAtPath: function (path) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath) || !literalStringMembers[subPath];
        }
        return true;
    },
    someReturnExpressionWhenCalledAtPath: function (path, _callOptions, predicateFunction, options) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath)
                || !literalStringMembers[subPath]
                || predicateFunction(options)(literalStringMembers[subPath].returns);
        }
        return true;
    },
    toString: function () { return '[[UNKNOWN STRING]]'; }
};
var returnsString = { value: { returns: UNKNOWN_LITERAL_STRING, callsArgs: null } };
var callsSecondArgReturnsString = { value: { returns: UNKNOWN_LITERAL_STRING, callsArgs: [1] } };
exports.UNKNOWN_OBJECT_EXPRESSION = {
    reassignPath: function () { },
    forEachReturnExpressionWhenCalledAtPath: function () { },
    getValue: function () { return exports.UNKNOWN_VALUE; },
    hasEffectsWhenAccessedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenAssignedAtPath: function (path) { return path.length > 1; },
    hasEffectsWhenCalledAtPath: function (path) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath) || !exports.objectMembers[subPath];
        }
        return true;
    },
    someReturnExpressionWhenCalledAtPath: function (path, _callOptions, predicateFunction, options) {
        if (path.length === 1) {
            var subPath = path[0];
            return VariableReassignmentTracker_1.isUnknownKey(subPath)
                || !exports.objectMembers[subPath]
                || predicateFunction(options)(exports.objectMembers[subPath].returns);
        }
        return true;
    },
    toString: function () { return '[[UNKNOWN OBJECT]]'; }
};
exports.objectMembers = assembleMemberDescriptions({
    hasOwnProperty: returnsBoolean,
    isPrototypeOf: returnsBoolean,
    propertyIsEnumerable: returnsBoolean,
    toLocaleString: returnsString,
    toString: returnsString,
    valueOf: returnsUnknown
});
exports.arrayMembers = assembleMemberDescriptions({
    concat: returnsArray,
    copyWithin: returnsArray,
    every: callsArgReturnsBoolean,
    fill: returnsArray,
    filter: callsArgReturnsArray,
    find: callsArgReturnsUnknown,
    findIndex: callsArgReturnsNumber,
    forEach: callsArgReturnsUnknown,
    includes: returnsBoolean,
    indexOf: returnsNumber,
    join: returnsString,
    lastIndexOf: returnsNumber,
    map: callsArgReturnsArray,
    pop: returnsUnknown,
    push: returnsNumber,
    reduce: callsArgReturnsUnknown,
    reduceRight: callsArgReturnsUnknown,
    reverse: returnsArray,
    shift: returnsUnknown,
    slice: returnsArray,
    some: callsArgReturnsBoolean,
    sort: callsArgReturnsArray,
    splice: returnsArray,
    unshift: returnsNumber
}, exports.objectMembers);
var literalBooleanMembers = assembleMemberDescriptions({
    valueOf: returnsBoolean
}, exports.objectMembers);
var literalNumberMembers = assembleMemberDescriptions({
    toExponential: returnsString,
    toFixed: returnsString,
    toLocaleString: returnsString,
    toPrecision: returnsString,
    valueOf: returnsNumber
}, exports.objectMembers);
var literalStringMembers = assembleMemberDescriptions({
    charAt: returnsString,
    charCodeAt: returnsNumber,
    codePointAt: returnsNumber,
    concat: returnsString,
    includes: returnsBoolean,
    endsWith: returnsBoolean,
    indexOf: returnsNumber,
    lastIndexOf: returnsNumber,
    localeCompare: returnsNumber,
    match: returnsBoolean,
    normalize: returnsString,
    padEnd: returnsString,
    padStart: returnsString,
    repeat: returnsString,
    replace: callsSecondArgReturnsString,
    search: returnsNumber,
    slice: returnsString,
    split: returnsArray,
    startsWith: returnsBoolean,
    substr: returnsString,
    substring: returnsString,
    toLocaleLowerCase: returnsString,
    toLocaleUpperCase: returnsString,
    toLowerCase: returnsString,
    toUpperCase: returnsString,
    trim: returnsString,
    valueOf: returnsString
}, exports.objectMembers);
function getLiteralMembersForValue(value) {
    switch (typeof value) {
        case 'boolean':
            return literalBooleanMembers;
        case 'number':
            return literalNumberMembers;
        case 'string':
            return literalStringMembers;
        default:
            return object_1.blank();
    }
}
exports.getLiteralMembersForValue = getLiteralMembersForValue;
function hasMemberEffectWhenCalled(members, memberName, callOptions, options) {
    return VariableReassignmentTracker_1.isUnknownKey(memberName)
        || !members[memberName]
        || (members[memberName].callsArgs
            && members[memberName].callsArgs.some(function (argIndex) { return callOptions.args[argIndex]
                && callOptions.args[argIndex].hasEffectsWhenCalledAtPath([], CallOptions_1.default.create({
                    withNew: false,
                    args: [],
                    callIdentifier: {} // make sure the caller is unique to avoid this check being ignored
                }), options.getHasEffectsWhenCalledOptions()); }));
}
exports.hasMemberEffectWhenCalled = hasMemberEffectWhenCalled;
function someMemberReturnExpressionWhenCalled(members, memberName, callOptions, predicateFunction, options) {
    return hasMemberEffectWhenCalled(members, memberName, callOptions, options)
        // if calling has no effect, memberName is a string and members[memberName] exists
        || predicateFunction(options)(members[memberName].returns);
}
exports.someMemberReturnExpressionWhenCalled = someMemberReturnExpressionWhenCalled;
