"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function getStarExcludes(_a) {
    var dependencies = _a.dependencies, exports = _a.exports;
    var starExcludes = new Set(exports.map(function (expt) { return expt.exported; }));
    if (!starExcludes.has('default'))
        starExcludes.add('default');
    // also include reexport names
    dependencies.forEach(function (_a) {
        var reexports = _a.reexports;
        if (reexports)
            reexports.forEach(function (reexport) {
                if (reexport.imported !== '*' && !starExcludes.has(reexport.reexported))
                    starExcludes.add(reexport.reexported);
            });
    });
    return starExcludes;
}
function system(chunk, magicString, _a) {
    var getPath = _a.getPath, t = _a.indentString, intro = _a.intro, outro = _a.outro;
    var _b = chunk.getModuleDeclarations(), dependencies = _b.dependencies, exports = _b.exports;
    var dependencyIds = dependencies.map(function (m) { return "'" + getPath(m.id) + "'"; });
    var importBindings = [];
    var starExcludes;
    var setters = [];
    var varOrConst = chunk.graph.varOrConst;
    dependencies.forEach(function (_a) {
        var imports = _a.imports, reexports = _a.reexports;
        var setter = [];
        if (imports) {
            imports.forEach(function (specifier) {
                importBindings.push(specifier.local);
                if (specifier.imported === '*') {
                    setter.push(specifier.local + " = module;");
                }
                else {
                    setter.push(specifier.local + " = module." + specifier.imported + ";");
                }
            });
        }
        if (reexports) {
            // bulk-reexport form
            if (reexports.length > 1 || reexports.length === 1 && reexports[0].imported === '*') {
                setter.push(varOrConst + " _setter = {};");
                // star reexports
                reexports.forEach(function (specifier) {
                    if (specifier.imported !== '*')
                        return;
                    // need own exports list for deduping in star export case
                    if (!starExcludes) {
                        starExcludes = getStarExcludes({ dependencies: dependencies, exports: exports });
                    }
                    setter.push("for (var _$p in module) {");
                    setter.push(t + "if (!_starExcludes[_$p]) _setter[_$p] = module[_$p];");
                    setter.push('}');
                });
                // reexports
                reexports.forEach(function (specifier) {
                    if (specifier.imported === '*')
                        return;
                    setter.push("_setter." + specifier.reexported + " = module." + specifier.imported + ";");
                });
                setter.push('exports(_setter);');
            }
            // single reexport
            else {
                reexports.forEach(function (specifier) {
                    setter.push("exports('" + specifier.reexported + "', module." + specifier.imported + ");");
                });
            }
        }
        setters.push(setter.join("\n" + t + t + t));
    });
    // function declarations hoist
    var functionExports = [];
    exports.forEach(function (expt) {
        if (expt.hoisted)
            functionExports.push("exports('" + expt.exported + "', " + expt.local + ");");
    });
    var starExcludesSection = !starExcludes ? '' :
        "\n" + t + varOrConst + " _starExcludes = { " + Array.from(starExcludes).join(': 1, ') + (starExcludes.size ? ': 1' : '') + " };";
    var importBindingsSection = importBindings.length ? "\n" + t + "var " + importBindings.join(', ') + ";" : '';
    var wrapperStart = "System.register([" + dependencyIds.join(', ') + "], function (exports, module) {\n" + t + "'use strict';" + starExcludesSection + importBindingsSection + "\n" + t + "return {" + (setters.length ? "\n" + t + t + "setters: [" + setters.map(function (s) { return "function (module) {\n" + t + t + t + s + "\n" + t + t + "}"; }).join(', ') + "]," : '') + "\n" + t + t + "execute: function () {\n\n" + (functionExports.length ? "" + t + t + t + functionExports.join("\n" + t + t + t) + '\n' : '');
    if (intro)
        magicString.prepend(intro);
    if (outro)
        magicString.append(outro);
    return magicString // TODO TypeScript: Awaiting PR
        .indent("" + t + t + t)
        .append("\n\n" + t + t + "}\n" + t + "};\n});")
        .prepend(wrapperStart);
}
exports.default = system;
